;* Title:		Micro-RX Processor
;* Version:		1.00
;* Target:		AT90S2313
;* Author:		wubblick@yahoo.com

.include "2313def.inc"
.include "macros.mac"

;***************************** Constantes: **********************************

.equ	Fclk	= 4000000	;uC  clock, Hz
.equ	Baud	= 19200		;baud rate
.equ	Tbas	=   10		;time base, mS
.equ	ARCnV	= 16		;slow autorepeat count
.equ	ARDel	= 80		;initial autorepeat delay, x Tbas
.equ	ARSlw	= 18		;slow autorepeat rate,	   x Tbas
.equ	ARFst	=  6		;fast autorepeat rate,	   x Tbas
.equ	DebV	=  3		;debounce delay,	   x Tbas

.equ	Fosc	=  12000000	;PLL clock, Hz
.equ	Fc	=     12500	;F reference, Hz
.equ	Fif	=  10700000	;F IF, Hz
.equ	MaxFrx	= 180000000	;max F RX tuning, Hz
.equ	MinFrx	= 140000000	;min F RX tuning, Hz
.equ	MaxFs	=    100000	;max F step, Hz
.equ	MinFs	= Fc		;min F step, Hz

.equ	Freq_0	= 144000000	;preset 0, Hz
.equ	Freq_1	= 144100000	;preset 1, Hz
.equ	Freq_2	= 144200000	;preset 2, Hz
.equ	Freq_3	= 144300000	;preset 3, Hz
.equ	Freq_4	= 144400000	;preset 4, Hz
.equ	Freq_5	= 144500000	;preset 5, Hz
.equ	Freq_6	= 144600000	;preset 6, Hz
.equ	Freq_7	= 144700000	;preset 7, Hz
.equ	Freq_8	= 144800000	;preset 8, Hz
.equ	Freq_9	= 144900000	;preset 9, Hz
.equ	Step_0	=     25000	;initial step, Hz

.equ T1V    = (((Fclk*Tbas)/100)+5)/10     ;CK/1
.equ UBRRV  = (((Fclk*10)/(16*Baud))-5)/10

;Keyboard scancodes:

.equ	K_NO	= 0x00		;no press
.equ	K_EX	= 0x01		;key EXIT code
.equ	K_DN	= 0x02		;key DOWN code
.equ	K_UP	= 0x04		;key UP code
.equ	K_EN	= 0x08		;key ENTER code

;Menu structure:

;[F  0]	<-> [FS 3]
;  |
;[P  1]
;  |
;[E  2]

.equ	MnuF	= 0		;edit Frequency
.equ	MnuP	= 1		;read Preset
.equ	MnuE	= 2		;preset save in Eeprom
.equ	MnuFS	= 3		;edit Frequency Step

;Serial link addresses:

.equ	Rx_Addr	= 0x00		;RX serial link address
.equ	Tx_Addr	= 0x00		;TX serial link address

;WAKE protocol constantes:

.equ	FEND	= 0xC0		;Frame END
.equ	FESC	= 0xDB		;Frame ESCape
.equ	TFEND	= 0xDC		;Transposed Frame END
.equ	TFESC	= 0xDD		;Transposed Frame ESCape

.equ	CRC_Init = 0xDE		;Initial CRC value
.equ	Frame	= 16		;maximum frame length

;Error codes:

.equ	Err_No	= 0x00		;no error
.equ	Err_Tx	= 0x01		;Rx/Tx error
.equ	Err_Bu	= 0x02		;device busy error
.equ	Err_Re	= 0x03		;device not ready error
.equ	Err_Pa	= 0x04		;parameters value error

;Command codes:

.equ	C_Nop	= 0x00		;No operation
.equ	C_Err	= 0x01		;Get Error code
.equ	C_Info	= 0x03		;Get Device Information
.equ	C_SetF	= 0x04		;Set Frequency
.equ	C_SetP	= 0x05		;Set Preset
.equ	C_GetP	= 0x06		;Get Preset

;***************************** Variables: ***********************************

.DSEG	;data segment (internal SRAM)

Frx:	.byte	4		;RX tuning frequency, Hz
Fs:	.byte	4		;frequency step, Hz
Pre:	.byte	4		;preset number
DRATIO:	.byte	2		;PLL D-ratio
MRATIO:	.byte	3		;PLL M-ratio (byte 3 not used)

KBD:	.byte 5			;keyboard data structure
.equ	Lc	= 0		;LastCode offset
.equ	Tc	= 1		;TempCode offset
.equ	ARCnt	= 2		;autorepeat counter offset
.equ	DebTM	= 3		;debounce timer offset
.equ	KeyTM	= 4		;key timer offset

Dig:	.byte 12		;display data (Digits 1..12, chars)
.equ	Dig0	= Dig+0		;8--------- char
.equ	Dig1	= Dig+1		;-8-------- char
.equ	Dig2	= Dig+2		;--8-------
.equ	Dig3	= Dig+3		;---8------
.equ	Dig4	= Dig+4		;----8-----
.equ	Dig5	= Dig+5		;-----8----
.equ	Dig6	= Dig+6		;------8---
.equ	Dig7	= Dig+7		;-------8--
.equ	Dig8	= Dig+8		;--------8-
.equ	Dig9	= Dig+9		;---------8
.equ	Dig10	= Dig+10	;---------- invisible
.equ	Dig11	= Dig+11	;---------- invisible

;WAKE protocol variables:

Rx_Str:	.byte Frame+6		;RX thread data structure
Tx_Str:	.byte Frame+6		;TX thread data structure

.equ	x_Pre	= 0		;pre-byte offset
.equ	x_Crc	= 1		;CRC offset
.equ	x_Ptr	= 2		;data pointer offset
.equ	x_Add	= 3		;address offset
.equ	x_Cmd	= 4		;command offset
.equ	x_Nbt	= 5		;number of bytes offset
.equ	x_Dat	= 6		;data offset

.equ	Rx_Dat	= Rx_Str+x_Dat	;RX data pointer
.equ	Tx_Dat	= Tx_Str+x_Dat	;TX data pointer

.ESEG	;EEPROM initial values

Enone:	.db 0			;address 0 not used
E_P0:	.db byte1(Freq_0)
	.db byte2(Freq_0)
	.db byte3(Freq_0)
	.db byte4(Freq_0)
	.db byte1(Freq_1)
	.db byte2(Freq_1)
	.db byte3(Freq_1)
	.db byte4(Freq_1)
	.db byte1(Freq_2)
	.db byte2(Freq_2)
	.db byte3(Freq_2)
	.db byte4(Freq_2)
	.db byte1(Freq_3)
	.db byte2(Freq_3)
	.db byte3(Freq_3)
	.db byte4(Freq_3)
	.db byte1(Freq_4)
	.db byte2(Freq_4)
	.db byte3(Freq_4)
	.db byte4(Freq_4)
	.db byte1(Freq_5)
	.db byte2(Freq_5)
	.db byte3(Freq_5)
	.db byte4(Freq_5)
	.db byte1(Freq_6)
	.db byte2(Freq_6)
	.db byte3(Freq_6)
	.db byte4(Freq_6)
	.db byte1(Freq_7)
	.db byte2(Freq_7)
	.db byte3(Freq_7)
	.db byte4(Freq_7)
	.db byte1(Freq_8)
	.db byte2(Freq_8)
	.db byte3(Freq_8)
	.db byte4(Freq_8)
	.db byte1(Freq_9)
	.db byte2(Freq_9)
	.db byte3(Freq_9)
	.db byte4(Freq_9)
E_FS:	.db byte1(Step_0)
	.db byte2(Step_0)
	.db byte3(Step_0)
	.db byte4(Step_0)

;************************ Global Register Variables: ************************

;r0 - lpm registers

.def	XSREG	= r1		;SREG store register

.def	Res1	= r4		;temporary registerd Res4:3:2:1
.def	Res2	= r5
.def	Res3	= r6
.def	Res4	= r7

.def	Buf1	= r8		;temporary registers Buf4:3:2:1
.def	Buf2	= r9
.def	Buf3	= r10
.def	Buf4	= r11

.def	Stp1	= r12		;temporary registers Stp4:3:2:1
.def	Stp2	= r13
.def	Stp3	= r14
.def	Stp4	= r15

.def	temp	= r16		;temporary register temp
.def	Cnt	= r17		;temporary register Cnt
.def	Menu	= r18		;menu item
.def	CMD	= r19
.def	Blink	= r20		;blink timer
  .equ	b2	= 0		;blink period = Tbas x 2
  .equ	b4	= 1		;blink period = Tbas x 4
  .equ	b8	= 2		;blink period = Tbas x 8
  .equ	b16	= 3		;blink period = Tbas x 16
  .equ	b32	= 4		;blink period = Tbas x 32
  .equ	b64	= 5		;blink period = Tbas x 64
  .equ	b128	= 6		;blink period = Tbas x 128
  .equ	b256	= 7		;blink period = Tbas x 256
.def	FLAGS	= r21
  .equ	UPDD	= 0		;display update flag
  .equ	UPDF	= 1		;frequency update flag
  .equ	NEWPR	= 2		;new pres flag

; r26,r27 used as X-register
; r28,r29 used as Y-register
; r30,r31 used as Z-register

;***** Port Definitions

.equ	DIRB	= 0b11101110	;Port B direction
.equ	PUPB	= 0b11110000	;Port B pull-ups

.equ	LDET	= PINB0		;Lock Detect
;	PB1	- not used
;	PB2	- not used
;	PB3	- not used
.equ	RETL	= PINB4		;Keyboard Return line
.equ	DLOAD	= PB5 		;Display Load line
.equ	DDATA	= PB6 		;Display Data line
.equ	DCLK	= PB7 		;Display Clock line
	
.equ	DIRD	= 0b00111010	;Port D direction
.equ	PUPD	= 0b00000111	;Port D pull-ups

.equ	RXD	= PIND0		;RS-232 RXD line
.equ	TXD	= PD1		;RS-232 TXD line
.equ	RDATA	= PIND2		;Received Data 
.equ	LE	= PD3		;PLL Latch Enable
.equ	DATA	= PD4		;PLL Serial Data
.equ	CLK	= PD5		;PLL Serial Clock
.equ	CDET	= PIND6		;Carrier Detect

;******************************* Macros: ************************************

.macro	ldxy
	ldi	XL,byte1(@0)
	ldi	XH,byte2(@0)
	ldi	YL,byte3(@0)
	ldi	YH,byte4(@0)
.endm

;************************** Interrupt Vectors: ******************************

.CSEG	;Code segment

.org	0
	rjmp	INIT		;reset handle
.org	OC1addr
	rjmp	TIMER1		;Timer 1 output compare interrupt handle

;******************* Program Execution Starts Here: *************************

INIT:	wdr			;watchdog reset
	ldi	temp,(1<<WDE) | (1<<WDP1)
	out	WDTCR,temp	;watchdog init (60 mS)	

	ldi	temp,RAMEND	;Locate stack
	out	SPL,temp

;Ports setup:
	
	ldi	temp,PUPB
	out	PORTB,temp	;init PORTB and on/off pullup
	ldi	temp,DIRB	
	out	DDRB,temp	;set PORTB direction
	
	ldi	temp,PUPD
	out	PORTD,temp	;init PORTD and on/off pullup
	ldi	temp,DIRD
	out	DDRD,temp	;set PORTD direction

;Variables init:

	clr	FLAGS		;clear flags
	ldi	YL,Pre
clrpre:	st	Y+,FLAGS	;Pre = 0
	cpi	YL,Pre+4
	brne	clrpre

;Periphery setup:

;USART (link to PC):

	ldi	temp,UBRRV
	out	UBRR,temp	;set baud rate

	ldi	temp,(1<<RXCIE) | (1<<TXCIE) | (1<<RXEN) | (1<<TXEN)
	out	UCR,temp	;UART init

;Timer 1 (system time):

	ldi	temp,0
	out	TCCR1A,temp	;OC disable, PWM disable
	ldi	temp,(1<<CTC1) | (1<<CS10)	
	out	TCCR1B,temp	;CTC, CK/1

	ldi	temp,byte2(T1V-1)
	out	OCR1AH,temp	
	ldi	temp,byte1(T1V-1)
	out	OCR1AL,temp

;Interrupts:

	ldi	temp,(1<<OCIE1A)
	out	TIFR,temp	;clear pending interrupts
	out	TIMSK,temp
	sei			;interrupts enable

;Subsystems init:

	ldi	Menu,MnuFS
	rcall	ReadE		;read Fs
	rcall	BufVd		;validate Fs
	rcall	BufSt		;save Fs

	ldi	Menu,MnuF
	rcall	ReadE		;Frx = Preset[0]
	rcall	BufVd		;validate Frx
	rcall	BufSt		;save Frx

	ldi	temp,0x0F	;temp <- BLK register address
	rcall	LCD_WA		;write address
	ldi	temp,0x0F	;temp <- 0x0F - enable bus
	rcall	LCD_WN		;write nibble

;Main loop:	
	
MAIN:	rcall	SCAN		;scan keyboard
	ldi	YL,KBD		;keyboard data structure base
	ldd	XL,Y+Lc	
	cp	XL,temp		;scancode = LastCode ?
	breq	Hold		;branch if same key
	
	ldd	XL,Y+Tc
	cp	XL,temp		;scancode = TempCode ?
	brne	NewP		;branch if new key
	
	ldd	XL,Y+DebTM
	tst	XL		;check debounce timer
	brne	Hold
	
	std	Y+Lc,temp	;LastCode <- scancode
	stbr	FLAGS,NEWPR	;set new press flag
	ldi	XH,ARDel	;autorepeat delay value
	ldi	XL,ARCnV
	rjmp	Stac		;go to store ARCnt
	
NewP:	std	Y+Tc,temp	;TempCode <- scancode
	ldi	XL,DebV
	std	Y+DebTM,XL	;debounce timer load
	
Hold:	clbr	FLAGS,NEWPR	;clear new press flag
	ldd	XL,Y+KeyTM
	tst	XL		;check key timer
	brne	Proc
	stbr	FLAGS,NEWPR	;set autorepeat flag
	ldd	XL,Y+ARCnt
	tst	XL
	breq	Fast		;fast autorepeat if count is over
	dec	XL		;dec autorepeat counter
	ldi	XH,ARSlw	;slow autorepeat rate
	
Stac:	std	Y+ARCnt,XL	;store autorepeat counter ARCnt
	rjmp	Stkt		;go to store KeyTM
	
Fast:	ldi	XH,ARFst	;fast autorepeat rate
Stkt:	std	Y+KeyTM,XH	;store key timer KeyTM
Proc:
	
;KBD+Lc	= true scancode,
;NEWPR	= 1 if pressed
;XL < ARCnV if autorepeat

;Process key functions:
	
	bbrc	FLAGS,NEWPR,NoPr
	ldd	temp,Y+Lc	;temp <- LastCode
	cpi	XL,ARCnV	;XL = ARCnV ?
	brne	Pro2		;skip EX, EN if autorepeat

Np:	cpi	temp,K_EX
	brne	Pro1
	rcall	Do_EX		;key EX processing
Pro1:	cpi	temp,K_EN
	brne	Pro2
	rcall	Do_EN		;key EN procesing

Pro2:	cpi	temp,K_UP
	brne	Pro3
	rcall	Do_UP		;key UP processing
Pro3:	cpi	temp,K_DN
	brne	NoPr
	rcall	Do_DN		;key DN processing
NoPr:

;Process PC commands:

	cpi	CMD,C_Err
	brne	Cmd1
	rcall	Do_Err		;Error comand
Cmd1:	cpi	CMD,C_Info
	brne	Cmd3
	rcall	Do_Info		;Info command
Cmd3:	cpi	CMD,C_SetF
	brne	Cmd4
	rcall	Do_SetF		;Set Frequency command
Cmd4:	cpi	CMD,C_SetP
	brne	Cmd5
	rcall	Do_SetP		;Set Preset command
Cmd5:	cpi	CMD,C_GetP
	brne	Cmd6
	rcall	Do_GetP		;Get Preset command
Cmd6:	

;Frequency update:

	bbrc	FLAGS,UPDF,nfu
	clbr	FLAGS,UPDF
	rcall	SetPLL		;set up PLL
nfu:

;Display copy update:

	bbrc	FLAGS,UPDD,ndu
	clbr	FLAGS,UPDD
	rcall	Disp		;update display copy
ndu:

;LDET and CDET polling:

	cpi	Menu,MnuF
	brne	npo
	ldi	YL,Dig
	ldi	temp,_F
	bbic	PINB,LDET,lok
	bbrs	Blink,b64,lok
	ldi	temp,BLANK
lok:	st	Y+,temp

	ldi	temp,_c
	bbic	PIND,CDET,cok
	ldi	temp,BLANK
cok:	st	Y+,temp
npo:		

;Load display:

	rcall	DispLd		;load display

;Watchdog restart:

	wdr			;watchdog restart
	rjmp 	MAIN		;main loop

;************************* Subroutines Area: ********************************

;-------------------------- Keyboard support: -------------------------------

;''''''''''''''''''''''''' Key "UP" processing: '''''''''''''''''''''''''''''

Do_UP:	rcall	BufLd		;load Buf
	rcall	StpLd		;load Stp
	cpi	Menu,MnuFS	;step edit ?
	brne	upns
	lsl	Buf1		;Buf * 2
	rol	Buf2
	rol	Buf3
	rol	Buf4		
	rjmp	upfs

upns:	rcall	Rem32		;Buf / Stp = Res.YH:YL:XH:XL
	sub	Buf1,XL		;Buf - Remainder
	sbc	Buf2,XH
	sbc	Buf3,YL
	sbc	Buf4,YH
	add	Buf1,Stp1	;Buff + Step
	adc	Buf2,Stp2
	adc	Buf3,Stp3
	adc	Buf4,Stp4
upfs:	rcall	BufVd		;validate value
	rcall	BufSt		;store Buf
	ret

;''''''''''''''''''''''''' Key "DN" processing: '''''''''''''''''''''''''''''
		
Do_DN:	rcall	BufLd		;load Buf
	rcall	StpLd		;load Stp
	cpi	Menu,MnuFS	;step edit ?
	brne	dnns
	lsr	Buf4		;Buf / 2
	ror	Buf3
	ror	Buf2
	ror	Buf1
	rjmp	dnfs
			
dnns:	rcall	Rem32		;Buf / Stp = Res.YH:YL:XH:XL
	mov	temp,XL
	or	temp,XH
	or	temp,YL
	or	temp,YH
	breq	rem0
	sub	Buf1,XL		;Buf - Remainder
	sbc	Buf2,XH
	sbc	Buf3,YL
	sbc	Buf4,YH
	rjmp	dnfs
rem0:	sub	Buf1,Stp1	;Buf - Step
	sbc	Buf2,Stp2
	sbc	Buf3,Stp3
	sbc	Buf4,Stp4
dnfs:	rcall	BufVd		;validate value
	rcall	BufSt
	ret
	
;''''''''''''''''''''''''' Key "EN" processing: '''''''''''''''''''''''''''''

Do_EN:	cpi	Menu,MnuF	;MnuF -> MnuFS
	brne	en1
	ldi	Menu,MnuFS	
	rjmp	enld
	
en1:	cpi	Menu,MnuP	;MnuP -> MnuF
	brne	en2
	ldi	Menu,MnuF
	stbr	FLAGS,UPDF	;set Frx update flag
	rjmp	enst

en2:	cpi	Menu,MnuE	;MnuE -> MnuF
	brne	en3
	rcall	SaveE		;save preset and Fs

en3:	ldi	Menu,MnuF	;MnuFS -> MnuF
	
enld:	rcall	BufLd
enst:	rcall	BufSt
	ret

;''''''''''''''''''''''''' Key "EX" processing: '''''''''''''''''''''''''''''

Do_EX:	cpi	Menu,MnuFS
	breq	Do_EN
	cpi	Menu,MnuE
	brne	ex1
	ldi	Menu,MnuF-1	;MnuE -> MnuF
ex1:	inc	Menu

	rcall	BufLd
	rcall	BufSt
	ret
	
;''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

;Load Buf:
;Input: Menu - mode
;Out:   Buf4:3:2:1 - data

BufLd:	cpi	Menu,MnuF
	brne	ldb1		;if (Menu == MnuF)
	ldi	XL,Frx		; Buf4:3:2:1 = Frx;
	rcall	LdBuf
	rjmp	ldbret
ldb1:	cpi	Menu,MnuFS
	brne	ldb2		;if (Menu == MnuFS)
	ldi	XL,Fs		; Buf4:3:2:1 = Fs;
	rcall	LdBuf
	rjmp	ldbret
ldb2:	ldi	XL,Pre		;if ((Menu == MnuP) || (Menu == MnuE))
	rcall	LdBuf		; Buf4:3:2:1 = Pre;
ldbret:	ret

;Load Buf4:3:2:1 from [X]:

LdBuf:	ld	Buf1,X+
	ld	Buf2,X+
	ld	Buf3,X+
	ld	Buf4,X+
	ret

;Load Stp:
;Input: Menu - mode
;Out:   Stp4:3:2:1 - data

StpLd:	cpi	Menu,MnuF
	brne	lds1		;if (Menu == MnuF)
	ldi	XL,Fs		; Stp4:3:2:1 = Fs;
	rcall	LdStp 
	rjmp	ldsret
lds1:	cpi	Menu,MnuFS
	breq	ldsret		;if ((Menu == MnuP) || (Menu == MnuE))
	clr	Stp4		; Stp4:3:2:1 = 1;
	clr	Stp3
	clr	Stp2
	clr	Stp1
	inc	Stp1		
ldsret:	ret

;Load Stp4:3:2:1 from [X]:

LdStp:	ld	Stp1,X+
	ld	Stp2,X+
	ld	Stp3,X+
	ld	Stp4,X+
	ret

;Validate value
;Input: Menu - mode
;       Buf4:3:2:1 - data
;Out:   Buf4:3:2:1 - valid data

BufVd:	cpi	Menu,MnuF
	brne	v1
	ldxy	MinFrx
	rcall	v_min
	ldxy	MaxFrx
	rcall	v_max
	rjmp	vret
v1:	cpi	Menu,MnuFS
	brne	v2
	ldxy	MinFs
	rcall	v_min
	ldxy	MaxFs
	rcall	v_max
	rjmp	vret
v2:	ldxy	0
	rcall	v_min
	ldxy	9
	rcall	v_max
vret:	ret

;Validate for minimum
;Input: YH:YL:XH:XL - min
;       Buf4:3:2:1
;Out:   Buf4:3:2:1

v_min:	cp	Buf1,XL
	cpc	Buf2,XH	
	cpc	Buf3,YL	
	cpc	Buf4,YH
	brge	minok
	mov	Buf1,XL
	mov	Buf2,XH
	mov	Buf3,YL
	mov	Buf4,YH
minok:	ret

;Validate for maximum
;Input: YH:YL:XH:XL - max
;       Buf4:3:2:1
;Out:   Buf4:3:2:1

v_max:	cp	Buf1,XL
	cpc	Buf2,XH	
	cpc	Buf3,YL	
	cpc	Buf4,YH
	brlt	maxok
	mov	Buf1,XL
	mov	Buf2,XH
	mov	Buf3,YL
	mov	Buf4,YH
maxok:	ret

;Store Buf:
;Input: Menu - mode
;       Buf4:3:2:1 - data
;Out:   Mem[Mode] - data

BufSt:	cpi	Menu,MnuF
	brne	stb1		;if (Menu == MnuF)
	ldi	XL,Frx
	rcall	StBuf		; { Frx = Buf4:3:2:1;
	stbr	FLAGS,UPDF	;   UPDF = 1; }
	rjmp	stba
stb1:	cpi	Menu,MnuFS
	brne	stb2		;if (Menu == MnuFS)
	ldi	XL,Fs
	rcall	StBuf		; Fs = Buf4:3:2:1;
	rjmp	stba
stb2:	cpi	Menu,MnuP
	brne	st3		;if (Menu == MnuP)
	ldi	XL,Pre
	rcall	StBuf		; { Pre = Buf4:3:2:1;
	rcall	ReadE		;   Buf4:3:2:1 = Preset[Pre];
	push	Menu
	ldi	Menu,MnuF
	rcall	BufVd		;   validate buffer; }
	pop	Menu
	rjmp	stba
st3:	cpi	Menu,MnuE
	brne	stba		;if (Menu == MnuE)
	ldi	XL,Pre
	rcall	StBuf		; { Pre = Buf4:3:2:1;
	push	Menu
	ldi	Menu,MnuF
	rcall	BufLd		;   Buf = Frx; }
	pop	Menu
stba:	stbr	FLAGS,UPDD	;UPDD = 1;
	ldi	temp,K_NO	;end of function processing;
	ret

;Store Buf4:3:2:1 to [X]:

StBuf:	st	X+,Buf1
	st	X+,Buf2
	st	X+,Buf3
	st	X+,Buf4
	ret

;Divide 32 / 32 = 32.32
;Buf4:3:2:1 / Stp4:3:2:1 = Res4:3:2:1.YH:YL:XH:XL
	
Rem32:	mov	Res1,Buf1
	mov	Res2,Buf2
	mov	Res3,Buf3
	mov	Res4,Buf4

;Res4:3:2:1 / Stp4:3:2:1 = Res4:3:2:1.YH:YL:XH:XL

Div32:	clr	XL		;clear remainder
	clr	XH
	clr	YL
	clr	YH
	clc			;clear carry
	ldi	Cnt,33		;init loop counter
	
d32u_1:	rol	Res1		;shift left dividend
	rol	Res2
	rol	Res3
	rol	Res4
	dec	Cnt		;decrement counter
	brne	d32u_2		;if done
	rjmp	done		;    jump out
	
d32u_2:	rol	XL		;shift dividend into remainder
	rol	XH
	rol	YL
	rol	YH
	sub	XL,Stp1		;remainder = remainder - divisor
	sbc	XH,Stp2
	sbc	YL,Stp3
	sbc	YH,Stp4	
	brcc	d32u_3		;if result negative
	add	XL,Stp1		;    restore remainder
	adc	XH,Stp2
	adc	YL,Stp3
	adc	YH,Stp4
	clc			;    clear carry to be shifted into result
	rjmp	d32u_1		;else
d32u_3:	sec			;    set carry to be shifted into result
	rjmp	d32u_1
done:	ret

;Scan local keyboard:
;Out: temp - scancode

SCAN:	rcall	scyc		;dummy scan (load 0bXXXX1111 in HC164)
	cbi	PORTB,DDATA	;DDATA <- 0 at first
scyc:	ldi	temp,0b00001000	;scan 4 buttons by 5 impulses
sclo:	cbi	PORTB,DCLK	;DCLK <- 0
	clc			;clc inside the loop for longer delay only
	sbis	PINB,RETL	;C <- ~RETL after /CLK
	sec
	sbi	PORTB,DCLK	;DCLK <- 1
	sbi	PORTB,DDATA	;DDATA <- 1 forever
	rol	temp		;C <- temp.7..temp.0 <- C
	brcc	sclo
	ret			;temp = scancode

;------------------------ PC commands processing: ---------------------------

;'''''''''''''''''''''' Command "C_Err" processing: '''''''''''''''''''''''''

;Tx_Dat[0] = Err_Tx - error code

Do_Err:
	ldi	XL,Tx_Dat
	ldi	temp,Err_Tx	;error code
	st	X+,temp
	ldi	Cnt,1		;N
	rjmp	Do_Cmd

;''''''''''''''''''''''' Command "C_Info" processing:''''''''''''''''''''''''

;Tx_Dat[] - Info string

Do_Info:
	table	Info
	ldi	YL,Tx_Dat
	clr	Cnt
Inf_C:	lpm
	st	Y+,r0
	adiw	ZH:ZL,1
	inc	Cnt		;N
	tst	r0
	brne	Inf_C
	rjmp	Do_Cmd

;''''''''''''''''''''''' Command "C_SetF" processing: '''''''''''''''''''''''

;Rx_Dat[0..3] - F [Hz]
;Tx_Dat[0]   - Error code = Err_No

Do_SetF:
	ldi	XL,Rx_Dat
	rcall	LdBuf
	ldi	Menu,MnuF	;MnuF
	rcall	BufVd
	rcall	BufSt

	ldi	Cnt,1		;N
	rjmp	Do_Cnd

;'''''''''''''''''''''' Command "C_SetP" processing: ''''''''''''''''''''''''

;Rx_Dat[0] - Preset number (0..9)
;Rx_Dat[1..4] - P [Hz]
;Tx_Dat[0] - Error code = Err_No

Do_SetP:
	ldi	XL,Rx_Dat
	ldi	Menu,MnuE
	rcall	Set_Pn
	rcall	LdBuf
	ldi	Menu,MnuF
	rcall	BufVd
	ldi	Menu,MnuE
	rcall	Do_En		;save preset, MnuF <- MnuE

	ldi	Cnt,1		;N
	rjmp	Do_Cnd

;'''''''''''''''''''''''' Command "C_GetP" processing: ''''''''''''''''''''''

;Rx_Dat[0] - Preset number (0..9)
;Tx_Dat[0] - Error code = Err_No
;Tx_Dat[1..4] - P [Hz]

Do_GetP:
	ldi	XL,Rx_Dat
	ldi	Menu,MnuP
	rcall	Set_Pn

	ldi	XL,Tx_Dat
	rcall	StBuf
	ldi	Menu,MnuF	;MnuF
	rcall	BufLd
	rcall	BufSt

	ldi	Cnt,5		;N

Do_Cnd:	ldi	XL,Tx_Dat
	ldi	temp,Err_No	;error code
	st	X+,temp
Do_Cmd:	ldi	ZL,Tx_Str
	std	Z+x_Nbt,Cnt	;load N
	std	Z+x_Cmd,CMD	;command
	rcall	UARTTX		;begin frame TX
	ldi	CMD,C_Nop	;end of command
	ret

;''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

;Set Preset number:

Set_Pn:	ld	Buf1,X+
	clr	Buf2
	clr	Buf3
	clr	Buf4
	rcall	BufVd
	rcall	BufSt
	ret

;--------------------------- EEPROM support: --------------------------------

;Read dword to Buf from the EEPROM:
;Input:	Menu - mode
;Out:   Buf4:3:2:1 - EEPROM data
	
ReadE:	rcall	EAddr		;XL - EEPROM address
	rcall	EE_Rdy
	mov	buf1,temp
	rcall	EE_Rdy
	mov	buf2,temp
	rcall	EE_Rdy
	mov	buf3,temp
	rcall	EE_Rdy
	mov	buf4,temp
	ret	

;Save dword to the EEPROM from Buf:
;Input:	Menu - mode
;       Buf4:3:2:1 - data

SaveE:	rcall	EAddr		;XL - EEPROM address
	mov	XH,Buf1
	rcall	EE_Sv
	mov	XH,Buf2
	rcall	EE_Sv
	mov	XH,Buf3
	rcall	EE_Sv
	mov	XH,Buf4
	rcall	EE_Sv
	cpi	XL,E_P0+4
	brne	sveret
	ldi	XL,E_FS		;save Fs
	ldi	YL,Fs
svfs:	ld	XH,Y+
	rcall	EE_Sv
	cpi	YL,Fs+4
	brne	svfs
sveret:	ret
	
;Make EEPROM address:
;Input: Menu - mode
;Out:   XL - EEPROM address

EAddr:	cpi	Menu,MnuFS
	brne	ea1
	ldi	XL,E_FS		;XL = EEPROM FS address	
	rjmp	earet
ea1:	ldi	XL,E_P0		;XL = EEPROM Preset[0] address
	cpi	Menu,MnuF
	breq	earet
	lds	temp,Pre
	add	XL,temp
	add	XL,temp
	add	XL,temp
	add	XL,temp		;XL = EEPROM Preset[Pre] address
earet:	ret
	
;Wait for EEPROM ready and read EEPROM:
;Input: XL - address
;Out:   temp - data, XL - next address

EE_Rdy: sbic	EECR,EEWE
	rjmp	EE_Rdy		;wait for EEPROM ready
	out	EEAR,XL		;EEPROM address
	sbi	EECR,EERE	;strobe
	in	temp,EEDR	;read EEPROM
	inc	XL		;EEPROM address inc
	ret

;Save byte to the EEPROM
;Input: XH - data byte
;       XL - address
;Out:   XL - next address

EE_Sv:	rcall	EE_Rdy		;temp - EEPROM data
	cp	XH,temp
	breq	skipEE
	out	EEDR,XH		;load data
	cli			;interrupts disable
	sbi 	EECR,EEMWE	;master write enable
	sbi	EECR,EEWE	;strobe
	sei			;interrupts enable
skipEE:	ret

;--------------------------- Display support: -------------------------------

;Display current editing value:
;Dig[0..1] <- StrT[Menu]
;Dig[1] <- preset number
;Dig[3..11] <- Frx or Fs
;Supress leading zeros
;Add point

Disp:	ldi	YL,Dig
	table	StrT
	mov	temp,Menu
	add	temp,Menu
	add	ZL,temp
	clr	temp
	adc	ZH,temp
	lpm
	st	Y+,r0		;menu char 1
	adiw	ZH:ZL,1
	lpm
	st	Y+,r0		;menu char 2
	ldi	temp,BLANK
	st	Y+,temp		;space

	cpi	Menu,MnuF
	breq	d1
	cpi	Menu,MnuFS
	breq	d1

	lds	temp,Pre+0
	sts	Dig+1,temp	;preset number

d1:	mov	Res1,Buf1
	mov	Res2,Buf2
	mov	Res3,Buf3
	mov	Res4,Buf4
	
;Res4:3:2:1 convert to BCD Dig[3..11]
	
bcd:	ldi	YL,Dig+2
	clr	temp
clrout: st	Y+,temp
	cpi	YL,Dig+12
	brne	clrout		;output array clear

	ldi	Cnt,32		;input bits count
	ldi	YH,Dig+2
	
hloop:	lsl	Res1		;input array shift left
	rol	Res2
	rol	Res3
	rol	Res4		
	ldi	YL,Dig+12
sloop:	ld	temp,-Y
	rol	temp
	subi	temp,-0x06	;temp+6, C=1
	sbrs	temp,4
	subi	temp,0x06	;temp-6, C=0
	andi	temp,0x0F
	st	Y,temp
	cpse	YL,YH		;YH = Dig+2
	rjmp	sloop
	dec	Cnt		;YL = Dig+2
	brne	hloop
	
;Supress leading zeros:
	
	ldi	temp,BLANK
slz:	ld	Cnt,Y
	tst	Cnt
	brne	notz
	st	Y+,temp
	cpi	YL,Dig+8
	brne	slz
	
;Setup point:

notz:	cpi	YL,Dig+6
	brsh	khzpnt
	ldi	YL,Dig+5
	rcall	setpnt
khzpnt:	ldi	YL,Dig+8
setpnt:	ld	Cnt,Y	
	ori	Cnt,0x80	;bit 7 - point
	st	Y,Cnt
	ret

;Load Display:
;Input: Dig[0..9]
	
DispLd:	ldi	temp,0x00	;temp <- 0x00 - SG1 address
	rcall	LCD_WA		;write address
	ldi	YL,Dig
ldmore:	ld	temp,Y+		;temp <- digit

;Write symbol from temp to LCD:

	bst	temp,7		;T <- temp.7
	andi	temp,0x7F	;temp.7 <- 0
	table	Font
	clr	r0
	add	ZL,temp
	adc	ZH,r0
	lpm
	mov	temp,r0
	
;Write data from r0,temp to LCD:

	rcall	LCD_WN		;write nibble from temp to LCD
	mov	temp,r0
	bld	temp,4		;bit 4 - segment H
	swap	temp
	rcall	LCD_WN		;write nibble from temp to LCD
	
	cpi	YL,Dig+10
	brne	ldmore		;repeat for 10 digits
	ret	

;Write nibble from temp to LCD:

LCD_WN:	andi	temp,0x0F	;mask unused bits
	ori	temp,0x10	;address=1
	rjmp	LCD_W5

;Write address from temp to LCD:

LCD_WA:	andi	temp,0x0F	;mask unused bits
LCD_W5:	ldi	Cnt,5		;write 5 bits to LCD
w5:	cbi	PORTB,DCLK	;DCLK <- 0
	cbi	PORTB,DDATA	;DDATA <- 0 or..
	sbrc	temp,4
	sbi	PORTB,DDATA	;DDATA <- 1
	rol	temp
	dec	Cnt
	sbi	PORTB,DCLK	;DCLK <- 1
	brne	w5
	cbi	PORTB,DLOAD	;DLOAD <- 0
	sbi	PORTB,DDATA
	sbi	PORTB,DLOAD	;DLOAD <- 1
	ret
		
;---------------------------- PLL support: ----------------------------------

;Set up PLL:
;make PLL DRATIO and MRATIO values and load PLL IC.
;Input: Frx - RX frequency, Hz
; const Fif - RX first IF frequency, Hz
; const Fosc - PLL oscillator frequency, Hz
; const Fc   - reference frequency, Hz
;
;Frx = Fvco - Fif
;Fvco = Fc * MRATIO
;Fc = Fosc / DRATIO

SetPLL:	

;DRATIO = Fosc / Fc

.equ	DR = Fosc/Fc

	ldi	XL,DRATIO
	ldi	temp,byte1(DR)
	st	X+,temp
	ldi	temp,byte2(DR)
	st	X+,temp

;Fvco = Frx + Fif

	ldi	XL,Frx
	ld	temp,X+
	subi	temp,byte1(-Fif)
	mov	Res1,temp
	ld	temp,X+
	sbci	temp,byte2(-Fif)
	mov	Res2,temp
	ld	temp,X+
	sbci	temp,byte3(-Fif)
	mov	Res3,temp
	ld	temp,X+
	sbci	temp,byte4(-Fif)
	mov	Res4,temp

;MRATIO = Fvco / Fc

	ldxy	(Fc)
	mov	Stp1,XL
	mov	Stp2,XH
	mov	Stp3,YL
	mov	Stp4,YH
	rcall	Div32		;Res4:3:2:1 / Stp4:3:2:1 = Res4:3:2:1
	ldi	XL,MRATIO
	st	X+,Res1
	st	X+,Res2
	st	X+,Res3

;load PLL IC:

	rcall	Wr_PLL
	ret

;Load Res4:3:2:1 from [X]:

LdRes:	ld	Res1,X+
	ld	Res2,X+
	ld	Res3,X+
	ld	Res4,X+
	ret

;Motorola MC12202 synthesizer load:
;
;Input: DRATIO =    8..16383  (2 bytes)
;       MRATIO = 1024..262143 (3 bytes)
;Out:        C = 1 if error
;
; F_r, F_p = F_osc / DRATIO
; F_vco = F_r, F_p * MRATIO
;
;DRATIO -> R, MRATIO -> P,N,A
;
;Note: since A should be less then N there are
;missing codes in 1024..4095 MRATIO range.
;Missing codes are indicated with C = 1 (and
;synthesizer is not loaded). In the same manner
;is indicated out of range for DRATIO or MRATIO.

Wr_PLL:	lds	XL,MRATIO+1
	lds	XH,MRATIO+2

;Check MRATIO for 1024..262143

	andi	XL,0xfc
	or	XL,XH
	brne	Mlk		;error if MRATIO < 1024 (400H)
	sec			;set error flag
	ret			;return to the caller
	
Mlk:	andi	XH,0xfc
	breq	Mgk		;error if MRATIO > 262143 (3FFFFH)
	sec			;set error flag
	ret			;return to the caller

;Check MRATIO and choose P
;Prescaler = 128 if MRATIO > 16383 (N > 127)
;Prescaler = 64  if MRATIO < 16384 (N < 128)
	
Mgk:	lds	XL,MRATIO+1
	lds	XH,MRATIO+2
	andi	XL,0xc0
	or	XL,XH
	in	temp,SREG
	bst	temp,1		;T <- Z

;Prescaler = 128 flag T = 0, Prescaler = 64 flag T = 1
;A_counter = (MRATIO MOD 128) for Prescaler = 128
;A_counter = (MRATIO MOD 64)  for Prescaler = 64
	
	lds	temp,MRATIO+0
	andi	temp,0x7f	;if P = 128 A_counter = MRATIO+0 & 0x7f
	brtc	P_128
	andi	temp,0x3f	;if P = 64  A_counter = MRATIO+0 & 0x3f

P_128:	lds	XH, MRATIO+2
	lds	XL, MRATIO+1
	lds	Cnt,MRATIO+0

	brts	P_64
	ror	XH		;if P = 128 MRATIO >> 2
	ror	XL
	ror	Cnt

P_64:	ror	XH		;if P = 64  MRATIO >> 1
	ror	XL
	ror	Cnt

	andi	Cnt,0xe0
	mov	XH,XL
	mov	XL,Cnt		;XH:XL = N18..N8,0,0,0,0,0
	push	XH		;_stack = 1 (N18..N11)
	push	XL		;_stack = 2 (N10..N8,0,0,0,0,0)

;temp = A_counter, XH:XL = N18..N8,0,0,0,0,0
;Check for A_counter < N

	ldi	Cnt,3
N_sh:	rol	XL		;N <<
	rol	XH
	brcs	Aok		;jump if N > A_counter (N > 255)
	dec	Cnt
	brne	N_sh	
	cp	temp,XH
	brlo	Aok		;error if A_counter >= N
	pop	XL		;_stack = 1
	pop	XH		;_stack = 0
	sec			;set error flag
	ret			;return to the caller
	
;Shift A_counter

Aok:	clc
	ldi	XL,0
	ldi	Cnt,2
A_sh:	ror	temp
	ror	XL
	dec	Cnt
	brne	A_sh
	mov	Cnt,XL		;temp:Cnt = A_counter >> 2

;Combine N18..N8 and A_counter

	pop	XL		;_stack = 1 (N10..N8,0,0,0,0,0)
	or	XL,temp
	push	XL		;_stack = 2 (N10..N8,A7..A3)
	push	Cnt		;_stack = 3 (A2,A1,0,0,0,0,0,0)

;_stack[1]:_stack[2]:_stack[3] ready to load PROGRAMMABLE DIVIDER

	lds	XL,DRATIO+0
	lds	XH,DRATIO+1
	
;Check DRATIO for 8..16383

	andi	XL,0xf8
	or	XL,XH
	brne	Dlk		;error if DRATIO < 8 (08H)
	pop	Cnt		;_stack = 2
	pop	XL		;_stack = 1
	pop	XH		;_stack = 0
	sec			;set error flag
	ret			;return to the caller

Dlk:	andi	XH,0xc0
	breq	Dgk		;error if DRATIO > 16383 (3FFFH)
	pop	Cnt		;_stack = 2
	pop	XL		;_stack = 1
	pop	XH		;_stack = 0
	sec			;set error flag
	ret			;return to the caller
	
;Shift DRATIO and combine it with SW (T flag)

Dgk:	lds	XL,DRATIO+0
	lds	XH,DRATIO+1
	sec
	rol	XL
	rol	XH
	bld	XH,7		;XH.7 <- T, XH:XL = SW,R14..R1,1

;XH:XL = SW,R14..R1,1 - ready to load PROGRAMMABLE REFERENCE DIVIDER

;Load PROGRAMMABLE REFERENCE DIVIDER

	mov	temp,XH
	ldi	Cnt,8		;bit count
	rcall	Ld_PLL		;load SW,R14..R8
	mov	temp,XL
	ldi	Cnt,8		;bit count
	rcall	Ld_PLL		;load R7..R1,1

	sbi	PORTD,LE
	nop
	cbi	PORTD,LE

;Load PROGRAMMABLE DIVIDER

	pop	XL		;_stack = 2 (A2,A1,0)
	pop	XH		;_stack = 1 (N10..N8,A7..A3)
	pop	temp		;_stack = 0 (N18..N11)
	ldi	Cnt,8		;bit count
	rcall	Ld_PLL		;load N18..N11
	mov	temp,XH
	ldi	Cnt,8		;bit count
	rcall	Ld_PLL		;load N10..N8,A7..A3
	mov	temp,XL
	ldi	Cnt,3		;bit count
	rcall	Ld_PLL		;load A2,A1,0

	sbi	PORTD,LE
	nop
	cbi	PORTD,LE

	clc			;clear error flag
	ret			;return to the caller

;Loads data to MC12202 via serial bus
;Input: temp - data, Cnt - bit count

Ld_PLL:	rol	temp
	brcc	DATA0
	sbi	PORTD,DATA	;DATA <- 1
	rjmp	DNE
DATA0:	cbi	PORTD,DATA	;DATA <- 0
	nop
DNE:	sbi	PORTD,CLK	;CLK  <- 1 
	nop
	cbi	PORTD,CLK	;CLK  <- 0
	dec	Cnt
	brne	Ld_PLL
	cbi	PORTD,DATA	;DATA <- 1
 	ret

;************************** Interrupts Handlers: ****************************

;Timer 1 Output Compare Interrupt (system time):

TIMER1:	in	XSREG,SREG
	push	YL
	push	YH
	inc	Blink		;Blink++;

	ldi	YL,KBD		;keyboard data structure base
	
	ldd	YH,Y+DebTM
	tst	YH		;check debounce timer		 
	breq	Tmr1
	dec	YH		;advance debounce timer
	std	Y+DebTM,YH

Tmr1:	ldd	YH,Y+KeyTM
	tst	YH		;check key timer		 
	breq	Tmr2
	dec	YH		;advance key timer
	std	Y+KeyTM,YH

Tmr2:	pop	YH
	pop	YL
	out	SREG,XSREG
	reti

;UART receive complete interrupt:

UARTRX:	push	temp
	in	temp,SREG	;save status register
	push	temp
	in	temp,UDR	;get byte
	sei			;enable nested interrupts
	push	Cnt
	push	ZL

	ldi	ZL,Rx_Str	;load RX data structure address

	cpi	temp,FEND
	brne	RX0
	std	Z+x_Pre,temp	;Rx_Pre = byte
	push	temp
	ldi	temp,CRC_Init
	std	Z+x_Crc,temp	;CRC init
	ldi	temp,x_Ptr
	std	Z+x_Ptr,temp	;pointer init
	pop	temp
	rjmp	RX12

RX0:	ldd	Cnt,Z+x_Ptr
	cpi	Cnt,x_Ptr	;if (Rx_Ptr == x_Ptr)
	brne	RX1
	rjmp	RX_RET		; Wait for FEND

RX1:	push	Cnt
	ldd	Cnt,Z+x_Pre	;load pre-byte
	std	Z+x_Pre,temp	;save new pre-byte
	mov	temp,Cnt
	pop	Cnt
	cpi	temp,FESC	;if (Rx_Pre == FESC)
	brne	RX3
	ldd	temp,Z+x_Pre
	cpi	temp,TFESC	; if (byte == TFESC)
	brne	RX2
	ldi	temp,FESC	;  byte = FESC
	rjmp	RX5
RX2:	cpi	temp,TFEND	;  else if (byte == TFEND)
	brne	RX_Err		;        byte = FEND
	ldi	temp,FEND	;	 else Rx_Err
	rjmp	RX5			

RX3:	ldd	temp,Z+x_Pre
	cpi	temp,FEND	;if (byte == FEND)
	breq	Rx_Err		; Rx_Err
RX4:	cpi	temp,FESC	;if (byte == FESC)
	breq	RX_RET		; skip

RX5:	cpi	Cnt,x_Add	;if (Rx_Ptr == x_Add)
	brne	RX7
	sbrc	temp,7		; if (byte.7 == 0)
	rjmp	RX6
	inc	Cnt		;  { inc R0
	std	Z+x_Ptr,Cnt	;    inc Rx_Ptr
	rjmp	RX11		;    save Tx_Cmd }
RX6:	clbr	temp,7
	tst	temp		; else if (!Addr) broadcast
	breq	RX12
	push	Cnt
	ldd	Cnt,Z+x_Add
	cp	temp,Cnt	;  else if (Addr == Rx_Add)
	pop	Cnt
	brne	RX10
	rjmp	RX12		;   match else new_frame

RX7:	cpi	Cnt,x_Cmd	;if (Rx_Ptr == x_Cmd)
	brne	RX8
	sbrc	temp,7		; if (byte.7 == 1)
	rjmp	Rx_Err		;  Rx_Err
	rjmp	RX11		;  else save Tx_Cmd	

RX8:	cpi	Cnt,x_Nbt	;if (Rx_Ptr == x_Nbt)
	brne	RX9
	cpi	temp,Frame	; if (Tx_Nbt < Frame)
	brlo	RX11		;  save Tx_Nbt
	ldi	temp,Frame	;  else { Tx_Nbt = Frame
	rjmp	RX11		;         save Tx_Nbt }

RX9:	push	temp
	ldd	temp,Z+x_Nbt
	subi	temp,-x_Nbt
	cp	temp,Cnt	;if (Rx_Ptr <= Rx_Nbt+x_Nbt)
	pop	temp
	brsh	RX11		; save Tx_Dat[] else

	ldd	Cnt,Z+x_Crc
	cp	temp,Cnt	;  if (byte != Rx_Crc) 
	brne	Rx_Err		;   Rx_Err
	ldd	CMD,Z+x_Cmd	;   else { CMD = Rx_Cmd
RX10:	ldi	temp,x_Ptr
	std	Z+x_Ptr,temp	;          Rx_Ptr = x_Ptr }
	rjmp	RX_RET

Rx_Err:	ldi	CMD,C_Err	;CMD = C_Err
	rjmp	RX10

RX11:	add	ZL,Cnt
	st	Z,temp		;save byte
	ldi	ZL,Rx_Str	;restore RX data structure address
RX12:	rcall	Do_Crc		;update CRC
	ldd	Cnt,Z+x_Ptr
	inc	Cnt
	std	Z+x_Ptr,Cnt	;inc Rx_Ptr

RX_RET:	pop	ZL
	pop	Cnt
	pop	temp
	out	SREG,temp	;restore status register
	pop	temp
	reti

;***** UART transmit complete interrupt

UARTTX:	sei			;enable nested interrupts
	push	temp
	in	temp,SREG	;save status register
	push	temp
	push	Cnt
	push	ZL

	ldi	ZL,Tx_Str	;load TX data structure address

	ldd	Cnt, Z+x_Ptr
	ldd	temp,Z+x_Nbt
	subi	temp,-x_Dat
	cp	temp,Cnt	;if (Tx_Ptr > Tx_Nbt+x_Dat)
	brsh	TX0
	ldi	temp,x_Ptr
	std	Z+x_Ptr,temp
	rjmp	TX_RET		; TX complete

TX0:	brne	TX1		;if (Tx_Ptr == Tx_Nbt+x_Dat)
	ldd	temp,Z+x_Crc	; byte = Tx_Crc
	rjmp	TX4

TX1:	cpi	Cnt,x_Ptr	;if (Tx_Ptr == x_Ptr)
	brne	TX2
	ldi	temp,CRC_Init	; { CRC init
	std	Z+x_Crc,temp
	ldi	temp,FEND	;   byte = FEND
	std	Z+x_Pre,temp	;   save pre-byte }
	rjmp	TX4

TX2:	cpi	Cnt,x_Add	;if (Tx_Ptr == x_Add)
	brne	TX3
	ldd	temp,Z+x_Add	; { byte = Tx_Add
	tst	temp
	brne	TX4		;   if (Tx_Add == 0)
	inc	Cnt		;   inc Cnt
	std	Z+x_ptr,Cnt	;   inc Tx_Ptr }

TX3:	add	ZL,Cnt
	ld	temp,Z		;get byte
	ldi	ZL,Tx_Str	;restore TX data structure address
TX4:	push	temp		;save uncoded byte

	cpi	Cnt,x_Add	;if (Tx_Ptr == x_Add)
	brne	TX5
	ori	temp,0x80	; byte = byte | 0x80

TX5:	cpi	Cnt,x_Ptr	;if (Tx_Ptr == x_Ptr)
	breq	TX9		; skip stuffing

TX6:	cpi	temp,FEND	;if (byte == FEND)
	brne	TX7
	ldd	temp,Z+x_Pre
	cpi	temp,FESC	; if (Tx_Pre == FESC)
	brne	TX8
	ldi	temp,TFEND	;  byte = TFEND
	rjmp	TX9		;   else...

TX7:	cpi	temp,FESC	;else if (byte == FESC)
	brne	TX9
	ldd	temp,Z+x_Pre
	cpi	temp,FESC	; if (Tx_Pre == FESC)
	brne	TX8
	ldi	temp,TFESC	;  byte = TFESC
	rjmp	TX9

TX8:	ldi	temp,FESC	;   else { byte = FESC
	out	UDR,temp	;          TX byte
	std	Z+x_Pre,temp	;          save pre-byte }
	pop	temp		;stack balance
	rjmp	TX_RET

TX9:	out	UDR,temp	;TX byte
	std	Z+x_Pre,temp	;save pre-byte
	pop	temp		;get uncoded byte
	rcall	Do_CRC		;update CRC
	ldd	Cnt,Z+x_Ptr
	inc	Cnt
	std	Z+x_Ptr,Cnt	;inc Tx_Ptr

TX_RET:	pop	ZL
	pop	Cnt
	pop	temp
	out	SREG,temp	;restore status register
	pop	temp
	reti

;Update CRC:
;Input: temp - data byte
;Z - pointer to Rx_Str or Tx_Str
;Out: [Z+x_Crc] - new CRC value

Do_CRC:	push	YL
	push	YH
	ldi	Cnt,8        	;Set to shift eight bits
	ldd	YH,Z+x_Crc
	
CRCL:	mov	YL,temp
	eor	YL,YH    	;Calculate DQIN xor CRCT0
	ror	YL          	;Move it to the carry
	brcc	ZERO        	;Skip if DQIN xor CRCT0 = 0
	ldi	YL,0x18
	eor	YH,YL    	;Update the CRC value
ZERO:	ror	YH          	;Position the new CRC
	lsr	temp           	;Position next bit in LSB
	dec	Cnt
	brne	CRCL		;Repeat for eight bits
	std	Z+x_Crc,YH
	pop	YH
	pop	YL
	ret              	;Return

;******************************* Tables: ************************************

;Font table:

Font:	     ;FCBHADEG    FCBHADEG
	.DB 0b11101110, 0b01100000	;0, 1
	.DB 0b00101111, 0b01101101	;2, 3
	.DB 0b11100001, 0b11001101	;4, 5
	.DB 0b11001111, 0b01101000	;6, 7
	.DB 0b11101111, 0b11101101	;8, 9
	.DB 0b10001011, 0b10101011	;F, P
	.DB 0b10001111, 0b00000111	;E, c
	.DB 0b00000001, 0b00000000	;-, blank
	
;Character codes table:

.equ	MINUS	=0x0E	;character "-" code
.equ	BLANK	=0x0F	;character "blank" code
.equ	__	=0x0F	;character "blank" code
.equ	_F	=0x0A	;character "F" code
.equ	_P	=0x0B	;character "P" code
.equ	_E	=0x0C	;character "E" code
.equ	_S	=0x05	;character "S" code
.equ	_c	=0x0D	;character "c" code

;String table:

StrT:	.DB _F,__,_P,__,_E,__,_F,_S

;Device Info:

Info:	.DB "uRX V1.00", 0

